//bestbud.cpp
//compila bestbud -c -omp
//Demora 1.5s
#include <cekeikon.h>

class BESTBUD {
  private:
    float gamma; int pz;
    Mat_<COR> Q, A;
    MATRIZ< vector<FLT> > IDrgb; 
    // Para cada pixel (l,c) de A, vetor de diferencas entre A[lA...lA+pz, cA...cA+pz]
    //   para Q (dividido em pz*pz)
    // Tem dimensao (A.rows-pz+1, A.cols-pz+1)
    Mat_<FLT> Dxy;
    int N; // Numero de pedacos pz*pz de Q
    FLT calculaIDrgb2(int lA, int cA, int lQ, int cQ);
    void calculaIDrgb();
    void calculaDxy();
    FLT calculaBBS(int l, int c);
  public:
    BESTBUD(Mat_<COR> Q, Mat_<COR> A, float _gamma=0.25, int _pz=3); // construtor
    Mat_<FLT> calcula(); // calcula BBS e devolve
};

FLT BESTBUD::calculaIDrgb2(int lA, int cA, int lQ, int cQ) {
  //Calcula distancia euclidiana de A[lA...lA+pz, cA...cA+pz]
  //  com Q[lQ...lQ+pz, cQ...cQ+pz], normalizando os niveis para 0 a 1.
  FLT f=0.0;
  for (int l=0; l<pz; l++)
    for (int c=0; c<pz; c++) {
      f+=elev2((A(lA+l,cA+c)[0]-Q(lQ+l,cQ+c)[0])/255.0);
      f+=elev2((A(lA+l,cA+c)[1]-Q(lQ+l,cQ+c)[1])/255.0);
      f+=elev2((A(lA+l,cA+c)[2]-Q(lQ+l,cQ+c)[2])/255.0);
    }
  return f/(3*pz*pz);
}

void BESTBUD::calculaIDrgb() {
  //Preenche todos os elementos (lA,cA) de IDrgb 
  //  com distancias de A(lA,cA) para Q
  IDrgb.resize(A.rows-pz+1, A.cols-pz+1);
  #pragma omp parallel for
  for (int lA=0; lA<IDrgb.rows; lA++)
    for (int cA=0; cA<IDrgb.cols; cA++) {
      IDrgb(lA,cA).reserve(N);
      for (int lQ=0; lQ<Q.rows; lQ+=pz)
        for (int cQ=0; cQ<Q.cols; cQ+=pz) {
          FLT f=calculaIDrgb2(lA,cA,lQ,cQ);
          IDrgb(lA,cA).push_back(f);
        }
    }
}

void BESTBUD::calculaDxy() {
  Dxy.create(N,N);
  #pragma omp parallel for
  for (int lA=0; lA<Q.rows; lA+=pz) {
    for (int cA=0; cA<Q.cols; cA+=pz) {
      int iA=lA/pz*Q.cols/pz+cA/pz;
      for (int lQ=0; lQ<Q.rows; lQ+=pz) 
        for (int cQ=0; cQ<Q.cols; cQ+=pz) {
          int iQ=lQ/pz*Q.cols/pz+cQ/pz;
          Dxy(iA,iQ)= norm( Point(cA,lA)-Point(cQ,lQ) );
        }
    }
  }
  Dxy=normaliza(Dxy);
}

BESTBUD::BESTBUD(Mat_<COR> _Q, Mat_<COR> _A, float _gamma, int _pz) {
  // Inicializa as variaveis. 
  // Calcula IDrgb para todos os pixels de A,
  // excluindo pz pixels inferiores e a direita.
  // Calcula Dxy
  gamma=_gamma; pz=_pz;
  Q=_Q.clone(); A=_A.clone();
  N=Q.total()/(pz*pz);

  if (!(Q.rows%pz==0 && Q.cols%pz==0 && A.rows%pz==0 && A.cols%pz==0)) 
    xerro1("Erro: As imagens Q e A devem ser divisiveis por pz");

  calculaIDrgb();
  calculaDxy();
}  

FLT BESTBUD::calculaBBS(int l, int c) {
  //Cria matriz Drgb de diferencas RGB em A(l,c), usando IDrgb
  //Calcula BBS em (l,c) usando Drgb
  if (!(0<=l && l<IDrgb.rows && 0<=c && c<IDrgb.cols)) 
    xerro1("Erro: Indice invalido");

  vector< vector<FLT>* > Drgb(N);  
  
  //Constroi matriz Drgb
  { int i=0;
    for (int lQ=0; lQ<Q.rows; lQ+=pz)
      for (int cQ=0; cQ<Q.cols; cQ+=pz) {
        Drgb[i]=&IDrgb(l+lQ,c+cQ);
        i++;
      }
  }
  
  //Constroi matriz D = Drgb + gamma*Dxy
  Mat_<FLT> D(N,N);
  for (int l=0; l<D.rows; l++)
    for (int c=0; c<D.cols; c++)
      D(l,c)=(*Drgb[l])[c]+gamma*Dxy(l,c);
      
  //Para cada linha de D, calcula o indice de coluna do menor elemento
  //Isto pode ser feito uma unica vez, quando construir IDrgb.
  vector<int> indl(N,0);
  for (int l=0; l<N; l++) {
    FLT minimo=FLT_MAX;
    for (int c=0; c<N; c++) {
      if ( D(l,c)<minimo ) {
        minimo=D(l,c);
        indl[l]=c;
      }
    }
  }
  
  //Para cada coluna de D, calcula o indice de linha do menor elemento
  vector<int> indc(N,0);
  for (int c=0; c<N; c++) {
    FLT minimo=FLT_MAX;
    for (int l=0; l<N; l++) {
      if ( D(l,c)<minimo ) {
        minimo=D(l,c);
        indc[c]=l;
      }
    }
  }

  //Conta o numero de best buddies.
  FLT soma=0;
  for (int i=0; i<N; i++) {
    int l=indl[i];
    if (!(0<=l && l<N)) xerro1("Erro: l fora do intervalo");
    int c=indc[l];
    if (c==i) soma=soma+1.0;
  }
    
  return soma;
}
  
Mat_<FLT> BESTBUD::calcula() {
  //Calcula BBS e devolve
  Mat_<FLT> BBS(A.rows-Q.rows+1, A.cols-Q.cols+1, 0.0);
  #pragma omp parallel for
  for (int l=0; l<BBS.rows; l++)
    for (int c=0; c<BBS.cols; c++) {
      BBS(l,c)=calculaBBS(l,c);
    }
  return BBS;
}

int _esquerda; int _acima;
Mat_<COR> fazDivisivel(Mat_<COR> a, int pz, int& esquerda=_esquerda, int& acima=_acima) {
  int sobral=a.rows%pz; acima=sobral/2; int abaixo=sobral-acima;
  int sobrac=a.cols%pz; esquerda=sobrac/2; int direita=sobrac-esquerda;
  
  Mat_<COR> b( a, Rect(esquerda, acima, a.cols-sobrac, a.rows-sobral) );
  Mat_<COR> d=b.clone();
  return d;
}

Mat_<FLT> impoeMinima(Mat_<FLT> reco, float& minima, float& maxima) {
  Mat_<FLT> sai=reco.clone();
  for (int i=0; i<sai.total(); i++) {
    sai(i)=(sai(i)-minima)/(maxima-minima);
  }
  return sai;
}

#define coloca(i) { int l; char st[256]; double x; l=256+borda-i; reta(d,l,128-96,l,138-96,0); \
                    x=i*(maxima-minima)/255.0+minima; sprintf(st,"%10.3f",x); putTxt(d,l-5,102-96,st); }

Mat_<GRY> escala(float minima, float maxima, string unidade)
// Desenha a escala degrade de 0 a 255, escrevendo unidade no canto inferior
{ int borda=10;
  Mat_<GRY> d(256+2*borda+16,256-60-96,255);
  for (int i=0; i<255; i++)
    for (int c=0; c<128-96; c++) {
      int l=256+borda-i;
      d(l,c)=i;
    }
  for (int i=0; i<256; i+=16) coloca(i);
  coloca(255);
  { int l=256+borda+16; 
    putTxt(d,l-5,142-96,"  "+unidade); 
  }

  return d;
}

int main(int argc, char** argv) {
  const char* keys =
    " curto| longo    |default| explicacao\n"
    "  -m  | -map     |       | Nome do arq de mapa de semelhanca\n"
    "  -g  | -gamma   | 0.05  | Peso da distancia xy\n"
    "  -p  | -pz      | 3     | Patch size\n"
    "    Ex: $bestbud q.ppm a.ppm sai.ppm -m=mapa.png -g=0.1 -p=4\n";

  if (argc<4) {
    printf("bestbud q.ppm a.ppm sai.ppm [opcoes]\n");
    printf("%s",keys);
    erro("Erro: Numero de argumentos invalido");
  }
  
  ArgComando cmd(argc,argv);
  Mat_<COR> Q; le(Q,cmd.getCommand(0));
  Mat_<COR> A; le(A,cmd.getCommand(1));
  string nomesai=cmd.getCommand(2);
  
  float gamma=cmd.getFloat("-g","-gamma",0.05);
  int pz=cmd.getInt("-p","-pz",3);

  string nomemap=cmd.getString("-m","-map");
  cmd.leuTodos();
  
  //recorta se nao for divisivel por pz
  Q=fazDivisivel(Q,pz);  
  int esquerda, acima;
  A=fazDivisivel(A,pz,esquerda,acima);  
  
  TimePoint t1=timePoint();
  BESTBUD bb(Q,A,gamma,pz); 
  TimePoint t2=timePoint();
  Mat_<FLT> f=bb.calcula();
  TimePoint t3=timePoint();
  
  f=normaliza(f);
  float minima=0.0, maxima=1.0;
  f=impoeMinima(f,minima,maxima);

  if (nomemap!="") {
    Mat_<GRY> g; converte(f,g);
    Mat_<GRY> esc=escala(minima,maxima,"BBS");
    Mat_<GRY> junto=grudaH(g,esc,10,255);
    Mat_<COR> sai=palAplic(junto,palKBGRW());
    imp(sai,nomemap);
  }
  
  Point maxLoc;
  minMaxLoc(f, NULL, NULL, NULL, &maxLoc);
  printf("Localizacao do centro na imagem original: l=%d c=%d\n",
    acima+Q.rows/2+maxLoc.y,esquerda+Q.cols/2+maxLoc.x);
  printf("Tempo de pre-eprocessamento=%6.2f calculo=%6.2f\n",timeSpan(t1,t2),timeSpan(t2,t3));
  rectangle(A, Rect(maxLoc,Q.size()), Scalar(0,0,255), 2);
  imp(A,nomesai);
}
  
  
