//Codigo em http://ipwithopencv.blogspot.com/2010/01/thin-plate-spline-example.html

/***************************************************************************
* Matthias Schmieder
***************************************************************************/
/**@file
@brief<b>Description: </b>
This is the header file of the CThinPlateSpline class. All functionality is wraped
and packed in this file. This class is used to compute non-rigid deformations using
the algorithm introduced by Bookstein in the late 80's. The original paper
"Principal Warps: Thin-Plate Splines and the Decomposition of Deformations" can be
found here http://cseweb.ucsd.edu/classes/sp03/cse252/bookstein.pdf

The alorithm was somehat improved but is nearly identical with the original proposed
in the original paper. There are plans to implement also some approximations and
more runtime efficient versions which were described here
http://cseweb.ucsd.edu/~sjb/pami_tps.pdf.

Up until now there is only the possibility to use the spline algorithm to warp
either forward or backward according to the energy evaluation of the corresponding
points given by the user.

<br> $Author: $ Matthias Schmieder
<br> $Date: $

***************************************************************************/

#ifndef CTHINPLATESPLINE_H_
#define CTHINPLATESPLINE_H_
using namespace cv;

/**
*****************************************************************************
The TPS_INTERPOLATION enumeration defines the usage of the class function.
Since interpolations is mostly done by back warping the images to prevent
un-defined pixels the the class will normally use BACK_WARP. Sometimes it
is neccessary to get the forward transformation which is done by using
FORWARD_WARP

@brief defines the warping type of the interpolation sheme
****************************************************************************/
enum TPS_INTERPOLATION
{
	FORWARD_WARP,                                   /**< forward transformation (optional)  */
	BACK_WARP                                               /**< back warp transform (standard)             */
};

/**
**************************************************************************
@brief CThinPlateSpline

This is the implementation of the original thin plate splien algorithm
introduce by Bookstein in 1989. This class helps you to use the algorithm
easy and efficient. See the function comments for more information.
***************************************************************************/
class CThinPlateSpline {
public:
	/**
	**************************************************************************
	CThinPlateSpline standard constructor.

	The standard constructor will do nothing else than set all control flags
	to the initial state. These contral flags are only used internally.
	***************************************************************************/
	CThinPlateSpline();


	/**
	**************************************************************************
	CThinPlateSpline constructor.

	One has the possibility to create an object and initializing it directly
	with corresponding point sets that will allow a direct use of the class.

	@param pS                       Reference to a STL vector that holds the points of
	the source image.

	@param pD                       Reference to a STL vector that holds the points of
	the target image.
	***************************************************************************/
	CThinPlateSpline(const std::vector<Point>& pS, const std::vector<Point>& pD);


	/**
	**************************************************************************
	CThinPlateSpline standard destructor.

	The standard destructor does nothing. All used datastructures are self-
	maintaining.
	***************************************************************************/

	~CThinPlateSpline();

	/**
	**************************************************************************
	The interpolation function is used to derive the position of a point
	transformed by the pre-computed spline coefficients. One has the possibility
	to warp forward or backward. The interpolation function determines for itself
	if the maps are suitable for forward/backward warps and recomputes them if
	necessary

	The spline is evaluated by computing the minimization of the energy function
	defined in Bookstein's original paper. The minimization function for the
	two dimensional problem is
	/f$ f(x,y) = a_1 + a_x*x + a_y * y + \sum_{i=1}^n w_i U(|P_i-(x,y)|)/f$

	As one can easily see the computation of one transformed pixel takes /f$O(n^2)/f$
	operations where n is the number of corresponding points.

	@param p                        Reference to the point that is going to be
	transformed/interpolated

	@param tpsInter defines if the point p is interpolated forward or
	backward


	@return                                 The function returns the interpolation result as
	cv::Point-structure
	***************************************************************************/
	Point interpolate(const Point& p, const TPS_INTERPOLATION tpsInter = BACK_WARP);

	/**
	**************************************************************************
	This function is used to add new corresponding points to the internal
	representation. When this function is called all internal states are reset
	so that the spline coefficients are recomputed.

	@param pS                       Reference to a point from the source image

	@param pD                       Reference to a point from the target image


	@return                                 void
	***************************************************************************/
	void addCorrespondence(const Point& pS, const Point& pD);

	/**
	**************************************************************************
	This is a combination of all function of this class. The function is used
	to directly transform an image according to the corresponding points
	provided earlier. If not already done, the function will compute the
	spline coefficients and then compute the transformation maps which will
	be used to transform the source image into it's destination using the
	cv::remap function. It is possible to choose the interpolation type as
	well as the transformation direction.

	@param src                                      Reference to the source image. Nothing will be done
	with this array other than reading values.

	@param dst                                      Reference the transformation destination. If dst
	was allocated before, it will be released and
	initialized with the correct size, depth etc.

	@param interpolation            Reference the transformation destination. If dst
	was allocated before, it will be released and
	initialized with the correct size, depth etc.
	All known interpolation function from the
	OpenCV library are possible: <br>
	INTER_NEAREST, <br>
	INTER_LINEAR, <br>
	INTER_AREA, <br>
	INTER_CUBIC, <br>
	INTER_LANCZOS4

	@param tpsInter                         This parameter defines the direction of the
	interpolation function. Possible values are: <br>
	FORWARD_WARP, <br>
	BACK_WARP

	@return                                 void
	***************************************************************************/
	void warpImage(const Mat& src,
		Mat& dst,
		double lambda = 0.001,
		const int interpolation = INTER_CUBIC,
		const TPS_INTERPOLATION tpsInter = BACK_WARP,
    const int borderType = BORDER_CONSTANT);

/*
Various border types, image boundaries are denoted with '|'
* BORDER_REPLICATE: aaaaaa|abcdefgh|hhhhhhh
* BORDER_REFLECT: fedcba|abcdefgh|hgfedcb
* BORDER_REFLECT_101: gfedcb|abcdefgh|gfedcba
* BORDER_WRAP: cdefgh|abcdefgh|abcdefg
* BORDER_CONSTANT: iiiiii|abcdefgh|iiiiiii with some specified 'i'
*/


	/**
	**************************************************************************
	This function is used to set the corresponding points directily. Instead
	of adding single point correspondences one can put hole vectors of points
	into this class, which are used to compute the spline coefficients.

	Make sure that both vectors have the same size. If this is not the case
	the computation will fail!

	@param pS                       Reference to a vector containing the points
	from the source image

	@param pD                       Reference to a vector containing the points
	from the source image

	@return                                 void
	***************************************************************************/
	void setCorrespondences(const std::vector<Point>& pS, const std::vector<Point>& pD);

	/**
	**************************************************************************
	The getMaps function is used to get your hands on the x- and y-values
	that are used for the interpolation. The maps contain the sub-pixel values
	which are used by cv::remap to interpolate each pixel.

	@param mapx                             sub-pixel x-coordinates

	@param mapy                             sub-pixel y-coordinates

	@return                                 void
	***************************************************************************/
	void getMaps(Mat& mapx, Mat& mapy);

	/**
	**************************************************************************
	If you want to get your hands on the maps to interpolate for yourself
	they have to be computed first. The following function will do this. The
	reason why getMaps() doesen't do this for itself is, that the getMaps
	function is supposed to only return the arrays without knowing if you want
	FORWARD- or BACK_WARP maps.

	@param dstSize                  size of the destination image. Here you can insert
	any size you like. The algorithm will simply
	evaluate the spline and look on the correct position
	in the array. If there is no value, the result will
	be black.

	@param mapx                             sub-pixel x-coordinates

	@param mapy                             sub-pixel y-coordinates

	@param tpsInter                 This parameter defines the direction of the
	interpolation function. Possible values are: <br>
	FORWARD_WARP, <br>
	BACK_WARP

	@return                                 void
	***************************************************************************/
	void computeMaps(const Size& dstSize,
		Mat_<float>& mapx,
		Mat_<float>& mapy,
		const TPS_INTERPOLATION tpsInter = BACK_WARP);

private:

	/**
	****************************************************************************
	\brief This is the main function of the thin plate spline algorithm.

	The function describes the surface on which a energy function will
	be minimized in the coefficient finding process. The function is defined
	as /f$ U(r) = -r^2 \log r^2/f$, where /f$ r = \sqrt(x^2+y^2)/f$ and is
	therefore the length of the vector connecting two corresponding points.

	@param  p1              First point (x,y) for the function computation
	@param  p2              First point (x,y) for the function computation

	@return                 the function will return a double value computed with the
	above mentioned function.

	****************************************************************************/
	double fktU(const Point& p1, const Point& p2);

	void computeSplineCoeffs(std::vector<Point>& iP,
		std::vector<Point>& iiP,
		double lambda,
		const TPS_INTERPOLATION tpsInter = BACK_WARP);

	Point interpolate_back_(const Point& p);
	Point interpolate_forward_(const Point& p);


	Mat_<double> cMatrix;
	Mat_<float> mapx;
	Mat_<float> mapy;
	std::vector<Point> pSrc;
	std::vector<Point> pDst;

	// FLAGS
	bool FLAG_COEFFS_BACK_WARP_SET;
	bool FLAG_COEFFS_FORWARD_WARP_SET;
	bool FLAG_MAPS_FORWARD_SET;
	bool FLAG_MAPS_BACK_WARP_SET;

};

#endif /* CTHINPLATESPLINE_H_ */

/*
* CThinPlateSpline.cpp
*
*  Created on: 24.01.2010
*      Author: schmiedm
*/

#include <vector>
#include <cv.h>
#include <cxcore.h>
using namespace cv;

CThinPlateSpline::CThinPlateSpline() {
	FLAG_MAPS_FORWARD_SET = false;
	FLAG_MAPS_BACK_WARP_SET = false;
	FLAG_COEFFS_BACK_WARP_SET = false;
	FLAG_COEFFS_FORWARD_WARP_SET = false;
}

CThinPlateSpline::CThinPlateSpline(const std::vector<Point>& pS, const std::vector<Point>& pD)
{
	if(pS.size() == pS.size())
	{
		pSrc = pS;
		pDst = pD;
	}

	FLAG_MAPS_FORWARD_SET = false;
	FLAG_MAPS_BACK_WARP_SET = false;
	FLAG_COEFFS_BACK_WARP_SET = false;
	FLAG_COEFFS_FORWARD_WARP_SET = false;
}

CThinPlateSpline::~CThinPlateSpline() {
}

void CThinPlateSpline::addCorrespondence(const Point& pS, const Point& pD)
{
	pSrc.push_back(pS);
	pDst.push_back(pD);

	// tell the class to recompute the coefficients if neccesarry
	FLAG_COEFFS_BACK_WARP_SET = false;
	FLAG_COEFFS_FORWARD_WARP_SET = false;
	FLAG_MAPS_FORWARD_SET = false;
	FLAG_MAPS_BACK_WARP_SET = false;

}

void CThinPlateSpline::setCorrespondences(const std::vector<Point>& pS, const std::vector<Point>& pD)
{
	pSrc = pS;
	pDst = pD;

	// tell the class to recompute the coefficients if neccesarry
	FLAG_COEFFS_BACK_WARP_SET = false;
	FLAG_COEFFS_FORWARD_WARP_SET = false;
	FLAG_MAPS_FORWARD_SET = false;
	FLAG_MAPS_BACK_WARP_SET = false;
}

double CThinPlateSpline::fktU(const Point& p1, const Point& p2)
{
	double r = pow(((double)p1.x - (double)p2.x), 2) + pow(((double)p1.y - (double)p2.y), 2);

	if (r == 0)
		return 0.0;
	else
	{
		r = sqrt(r); // vector length
		double r2 = pow(r, 2);

		//return (r2 * log(r2));
		return (r2 * log(r));  //Segundo um comentario no site ????
	}
}

void CThinPlateSpline::computeSplineCoeffs(std::vector<Point>& iPIn, std::vector<Point>& iiPIn, double lambda,const TPS_INTERPOLATION tpsInter)
{

	std::vector<Point>* iP = NULL;
	std::vector<Point>*	iiP = NULL;

	if(tpsInter == FORWARD_WARP)
	{
		iP = &iPIn;
		iiP = &iiPIn;

		// keep information which coefficients are set
		FLAG_COEFFS_BACK_WARP_SET = true;
		FLAG_COEFFS_FORWARD_WARP_SET = false;
	}
	else if(tpsInter == BACK_WARP)
	{
		iP = &iiPIn;
		iiP = &iPIn;

		// keep information which coefficients are set
		FLAG_COEFFS_BACK_WARP_SET = false;
		FLAG_COEFFS_FORWARD_WARP_SET = true;
	}

	//get number of corresponding points
	int dim = 2;
	int n = iP->size();

	//Initialize mathematical datastructures
	Mat_<double> V(dim,n+dim+1,0.0);
	Mat_<double> P(n,dim+1,1.0);
	Mat_<double> K = (K.eye(n,n)*lambda);
	Mat_<double> L(n+dim+1,n+dim+1,0.0);

	// fill up K und P matrix
	std::vector<Point>::iterator itY;
	std::vector<Point>::iterator itX;

	int y = 0;
	for (itY = iP->begin(); itY != iP->end(); ++itY, y++) {
		int x = 0;
		for (itX = iP->begin(); itX != iP->end(); ++itX, x++) {
			if (x != y) {
				K(y, x) = (double)fktU(*itY, *itX);
			}
		}
		P(y,1) = (double)itY->y;
		P(y,2) = (double)itY->x;
	}

	Mat Pt;
	transpose(P,Pt);

	// insert K into L
	Rect range = Rect(0, 0, n, n);
	Mat Lr(L,range);
	K.copyTo(Lr);


	// insert P into L
	range = Rect(n, 0, dim + 1, n);
	Lr = Mat(L,range);
	P.copyTo(Lr);

	// insert Pt into L
	range = Rect(0,n,n,dim+1);
	Lr = Mat(L,range);
	Pt.copyTo(Lr);

	// fill V array
	std::vector<Point>::iterator it;
	int u = 0;

	for(it = iiP->begin(); it != iiP->end(); ++it)
	{
		V(0,u) = (double)it->y;
		V(1,u) = (double)it->x;
		u++;
	}

	// transpose V
	Mat Vt;
	transpose(V,Vt);

	cMatrix = Mat_<double>(n+dim+1,dim,0.0);

	// invert L
	Mat invL;
	invert(L,invL,DECOMP_LU);

	//multiply(invL,Vt,cMatrix);
	cMatrix = invL * Vt;

	//compensate for rounding errors
	for(int row = 0; row < cMatrix.rows; row++)
	{
		for(int col = 0; col < cMatrix.cols; col++)
		{
			double v = cMatrix(row,col);
			if(v > (-1.0e-006) && v < (1.0e-006) )
			{
				cMatrix(row,col) = 0.0;
			}
		}
	}
}


Point CThinPlateSpline::interpolate_forward_(const Point& p)
{
	Point2f interP;
	std::vector<Point>* pList = &pSrc;

	int k1 = cMatrix.rows - 3;
	int kx = cMatrix.rows - 2;
	int ky = cMatrix.rows - 1;

	double a1 = 0, ax = 0, ay = 0, cTmp = 0, uTmp = 0, tmp_i = 0, tmp_ii = 0;

	for (int i = 0; i < 2; i++) {
		a1 = cMatrix(k1,i);
		ax = cMatrix(kx,i);
		ay = cMatrix(ky,i);

		tmp_i = a1 + ax * p.y + ay * p.x;
		tmp_ii = 0;

		for (int j = 0; j < (int)pSrc.size(); j++) {
			cTmp = cMatrix(j,i);
			uTmp = fktU( (*pList)[j], p);

			tmp_ii = tmp_ii + (cTmp * uTmp);
		}

		if (i == 0) {
			interP.y = (double)(tmp_i + tmp_ii);
		}
		if (i == 1) {
			interP.x = (double)(tmp_i + tmp_ii);
		}
	}

	return interP;
}
Point CThinPlateSpline::interpolate_back_(const Point& p)
{
	Point2f interP;
	std::vector<Point>* pList = &pDst;

	int k1 = cMatrix.rows - 3;
	int kx = cMatrix.rows - 2;
	int ky = cMatrix.rows - 1;

	double a1 = 0, ax = 0, ay = 0, cTmp = 0, uTmp = 0, tmp_i = 0, tmp_ii = 0;

	for (int i = 0; i < 2; i++) {
		a1 = cMatrix(k1,i);
		ax = cMatrix(kx,i);
		ay = cMatrix(ky,i);

		tmp_i = a1 + ax * p.y + ay * p.x;
		tmp_ii = 0;

		for (int j = 0; j < (int)pSrc.size(); j++) {
			cTmp = cMatrix(j,i);
			uTmp = fktU( (*pList)[j], p);

			tmp_ii = tmp_ii + (cTmp * uTmp);
		}

		if (i == 0) {
			interP.y = (double)(tmp_i + tmp_ii);
		}
		if (i == 1) {
			interP.x = (double)(tmp_i + tmp_ii);
		}
	}

	return interP;
}

Point CThinPlateSpline::interpolate(const Point& p, const TPS_INTERPOLATION tpsInter)
{
	if(tpsInter == BACK_WARP)
	{
		return interpolate_back_(p);
	}
	else if(tpsInter == FORWARD_WARP)
	{
		return interpolate_forward_(p);
	}
	else
	{
		return interpolate_back_(p);
	}

}

void CThinPlateSpline::warpImage(const Mat& src, Mat& dst, double lambda,
  const int interpolation, const TPS_INTERPOLATION tpsInter, const int borderType)
{
	Size size = src.size();
	dst = Mat(size,src.type());

	// only compute the coefficients new if they weren't already computed
	// or there had been changes to the points
	if(tpsInter == BACK_WARP && !FLAG_COEFFS_BACK_WARP_SET)
	{
		computeSplineCoeffs(pSrc,pDst,lambda,tpsInter);
	}
	else if(tpsInter == FORWARD_WARP && !FLAG_COEFFS_FORWARD_WARP_SET)
	{
		computeSplineCoeffs(pSrc,pDst,lambda,tpsInter);
	}

	computeMaps(size,mapx,mapy);

	remap(src,dst,mapx,mapy,interpolation,borderType);
}

void CThinPlateSpline::getMaps(Mat& mx, Mat& my)
{
	mx = mapx;
	my = mapy;
}

void CThinPlateSpline::computeMaps(const Size& dstSize, Mat_<float>& mx, Mat_<float>& my,const TPS_INTERPOLATION tpsInter)
{
	mx = Mat_<double>(dstSize);
	my = Mat_<double>(dstSize);

	Point p(0, 0);
	Point_<double> intP(0, 0);

	for (int row = 0; row < dstSize.height; row++) {
		for (int col = 0; col < dstSize.width; col++) {
			p = Point(col, row);
			intP = interpolate(p,tpsInter);
			mx(row, col) = intP.x;
			my(row, col) = intP.y;
		}
	}

	if(tpsInter == BACK_WARP)
	{
		FLAG_MAPS_BACK_WARP_SET = true;
		FLAG_MAPS_FORWARD_SET = false;
	}
	else if(tpsInter == FORWARD_WARP)
	{
		FLAG_MAPS_BACK_WARP_SET = false;
		FLAG_MAPS_FORWARD_SET = true;
	}
}

